"""create the coordinates for the elastic search polygon instance"""

cpdef list convex_array(list latitudes, list longitudes, bint reverse_param):

    """create convex array (encapsulated) for elasticsearch polygon geoshape query"""

    cdef list coordinates = []

    if reverse_param:
        coordinates = [f"{long}, {lat}" for lat, long in zip(latitudes, longitudes)]
    else:
        coordinates = [f"{lat}, {long}" for lat, long in zip(latitudes, longitudes)]

    return coordinates


"""throttle confidence mapping code"""

# noinspection PyTypeChecker
cpdef float confidence_map(dict scale, list index, float distance):

    """map the confidence to a certainty value"""

    # define array length
    cdef int length = len(index)
    cdef int i = 0
    cdef list stack = []
    cdef float confidence = 0.0

    # loop through array and add to stack if condition satisfied
    for i in range(1, length + 1):
        # if distance is larger than the previous index item, add to stack
        if distance >= index[i - 1]:
            # insert in stack
            stack.append(scale[index[i]])
        else:
            continue

    # pop from the stack first
    try:
        confidence = stack.pop()
    # if stack is empty, it means it is the most accurate 1.0 mapping
    except IndexError:
        # return the mapping
        confidence = scale[index[0]]

    return confidence


"""state-machine certainty algorithms"""
from state_machine import State, Event, acts_as_state_machine, after, InvalidStateTransition
from typing import Union, Callable
from time import perf_counter


# noinspection PyCompatibility
@acts_as_state_machine
class CertaintySelection:

    """implement the state-machine for certainty logic"""

    # defined states
    length = State(initial=True)
    processed = State()

    # defined events
    converted = Event(from_states=length, to_state=processed)

    def __init__(self, scale_index: dict, distance: Union[float, int], algorithm: Callable):
        self.scale = scale_index
        # noinspection PyTypeChecker
        self.index: list = list(self.scale.keys())
        self.distance = distance
        # self.stack = []  LifoQueue(maxsize=5)
        self.confidence: float = 0.0
        self.algorithm = algorithm

    @after('converted')
    def length_conv(self):

        """get the confidence for a given distance on the index/scale"""
        self.confidence = self.algorithm(self.scale, self.index, self.distance)


def state_controller(scale: dict, distance: Union[float, int], algorithm: Callable):
    """controls the states"""

    # create state-machine object
    state_mach = CertaintySelection(scale, distance, algorithm)

    # control state transitioning
    try:
        # noinspection PyCallingNonCallable
        state_mach.converted()
        return state_mach.confidence
    except InvalidStateTransition or TypeError:
        raise BrokenPipeError


"""scale = {0.0: 1.0, 1.0: 1.0, 10.0: 0.9, 100.0: 0.8, 250.0: 0.7, 1000.0: 0.6, 1000.99: 0.5}
start = perf_counter()
print(state_controller(scale, 101))
print(perf_counter() - start)"""


# ------------------------- TEST FUNCTIONS ---------------------------

# noinspection PyCompatibility
def test1(distance, confidence=0.0):

    """get the confidence for a given distance on the index/scale"""

    # set boolean flag
    flag: bool = False

    # define array length
    length = len(index)

    # run iteration over conditional array options
    while flag is False:
        for i in range(1, length - 1):
            if index[i - 1] < distance < index[i + 1]:
                confidence = confidence_index[index[i]]
                flag = True
            else:
                continue

        if not flag:
            print("in not flag")
            if distance > index[-1]:
                confidence = confidence_index[index[-1]]
                flag = True
            else:
                confidence = confidence_index[index[0]]
                flag = True

    return confidence


def test2(distance, confidence=0.0):
    stack = []
    for i in range(1, len(index2)):
        if distance > index2[i-1]:
            stack.append(confidence_index2[index2[i]])
        else:
            continue

    # print(stack)
    try:
        return stack.pop()
    except IndexError:
        return confidence_index2[index2[-1]]



"""backend functionality to PeekGeo"""
from abc import ABCMeta, abstractmethod
from pydantic import BaseModel
import sys
import os
sys.path.append("app/")
import pymongo
from pymongo.errors import OperationFailure
from typing import List, Union, Any, Type, Dict, Tuple, Optional, Callable, Iterable
import ssl
from bson import BSON

# elasticsearch imports
from elasticsearch import Elasticsearch
from elasticsearch.exceptions import NotFoundError, RequestError
from configparser import ConfigParser

# scientific packages
import numpy as np
from scipy.spatial import ConvexHull

# common built-ins
from places_geocode.app.generator_BSON import handler_func
# noinspection PyCompatibility
import asyncio
# noinspection PyCompatibility
from concurrent.futures import ThreadPoolExecutor, as_completed
import time
from math import pi, sin, cos, sqrt, atan2, radians
import sys
from threading import Thread
from collections import OrderedDict

# data model imports
from places_geocode.app.models import density_models, common_models, reverse_models, forward_models, generator_model, certainty_models, \
    unit_conversion_model, correction_parse_models

# cython modules import
from ctypes import *

path = os.path.dirname(os.path.realpath(__file__))

try:
    spreadsheet_c = cdll.LoadLibrary(os.path.abspath("spreadsheet_c.cpython-38-darwin.so"))
    reverse_spreadsheet = cdll.LoadLibrary(os.path.abspath("reverse_spreadsheet.cpython-38-darwin.so"))
    TFIDF_max = cdll.LoadLibrary(os.path.abspath("TFIDF_max.cpython-38-darwin.so"))
    null_parameters = cdll.LoadLibrary(os.path.abspath("null_parameters.cpython-38-darwin.so"))
    certainty_search = cdll.LoadLibrary(os.path.abspath("certainty_search.cpython-38-darwin.so"))
    radius_arr = cdll.LoadLibrary(os.path.abspath("radius_arr.cpython-38-darwin.so"))
    TFIDF_max_elastic = cdll.LoadLibrary(os.path.abspath("TFIDF_max_elastic.cpython-38-darwin.so"))
    print('finished backend try')
except:
    from places_geocode.app.CythonOperations import spreadsheet_c, reverse_spreadsheet, TFIDF_max, null_parameters, TFIDF_max_elastic, certainty_search, radius_arr

# ----------------------DataClasses Here---------------------


# noinspection PyTypeChecker
class BuildObjs:
    """builder for the mongo objects"""

    def __init__(self):
        self.set_up_objs = dict(database=None, common_model=None, client=None)

    @staticmethod
    def _create_max_size_model(max_size: int):
        """add max_size_model to the common models object dataclass"""
        common_model = common_models.ObjModel(maxSize=max_size)
        return common_model

    @staticmethod
    def _create_client(max_size: int):
        """make the mongo client"""
        client = pymongo.MongoClient("mongodb+srv://martin_mashalov:Barca2004!@locationpoints.wi1dk.mongodb.net/myFirstDatabase?retryWrites=true&w=majority&authSource=admin")
        return client

    @staticmethod
    def _create_connection(database, collection, client):
        """create the database connection to the collection specified"""
        db = client[database]
        collection = db[collection]

        return db, collection

    def set_max_size(self, *args):
        """set the max_size in the dataclass model"""

        # set the maxsize and create common model
        self.set_up_objs['common_model'] = self._create_max_size_model(args[0])

    def set_client(self):
        """set the client object"""

        # get the client object
        self.set_up_objs['client'] = self._create_client(self._create_max_size_model(100).maxSize)

    def build_connection(self, database, collection):
        """build the database connection with the builder pattern"""

        # get the database, collection objects, using client
        self.set_up_objs['database'], self.set_up_objs[collection] = self._create_connection(database, collection,
                                                                                             self.set_up_objs['client'])


def DirectorGeneralObjs(database: str, collections: Tuple[str], *args) -> Dict:
    """direct the building process for mongodb object creation"""
    builder_direct = BuildObjs()

    # set for building mongo object
    builder_direct.set_max_size(args[0])
    builder_direct.set_client()
    builder_direct.build_connection(database, collections[0])
    builder_direct.build_connection(database, collections[1])
    return builder_direct.set_up_objs


# noinspection PyTypeChecker
class BuildElasticSearch:
    """build elasticsearch objects"""

    def __init__(self):
        self.elastic_settings = ElasticModel()
        self.set_up_objs = dict(connection=None)
        self.config = ConfigParser()
        self.config.read('places_geocode/app/example.ini')

    def create_connection(self):
        """create the connection for locations"""
        elastic_con = Elasticsearch(
            cloud_id=self.config['ELASTIC']['cloud_id'],
            http_auth=(self.config['ELASTIC']['user'], self.config['ELASTIC']['password'])
        )
        self.set_up_objs['connection'] = elastic_con


def ElasticBuildDirector(*opt):
    """elasticsearch connection/object builder"""
    builder = BuildElasticSearch()
    builder.create_connection()
    return builder.set_up_objs


class ThreadWithReturnValue(Thread):
    """manual implementation of threading for multiple functions made simple"""

    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs=None, Verbose=None):
        Thread.__init__(self, group, target, name, args, kwargs)
        if kwargs is None:
            kwargs = {}
        self._return = None

    def run(self):
        """run the thread functionality"""
        print(type(self._target))
        if self._target is not None:
            self._return = self._target(*self._args,
                                        **self._kwargs)

    def join(self, *args):
        """thread join method"""
        Thread.join(self, *args)
        return self._return


"""--------------------------------Forward Geocoding Data Model------------------------------------"""


# noinspection PyCompatibility
class LocalDataClass(BaseModel):
    """basic parameters for the forward geocoding functionality"""
    dflib_hyperparameter: Optional[int] = 1
    similarity_threshold: Optional[float] = 0.5
    batch_size: Optional[int] = 125
    small_batch_maximum: Optional[int] = 4
    spreadsheet_batch_size: Optional[int] = 10000
    optimal_radius: Optional[int] = 10
    common_streets: Optional[List[str]] = ['road', 'street', 'avenue', 'drive', 'lane', 'route', 'court', 'west',
                                           'east', 'north', 'south', 'place', 'state', 'boulevard', 'county', 'way',
                                           'square', 'pl', 'dr', 'bl']
    forward_geocode_street_method: Optional[Callable] = lambda street: street.upper()
    confidence_index: Optional[Dict[float, float]] = {0.0: 1.0, 1.0: 1.0, 10.0: 0.9, 100.0: 0.8, 250.0: 0.7,
                                                      1000.0: 0.6, 1000.99: 0.5}
    confidence_algorithm: Optional[Callable] = certainty_search
    default_autocomplete_algorithm: Optional[Callable] = None
    radius_delta_null: Optional[float] = 0.001

    # assign builder
    builder_utility: Optional[Callable] = ElasticBuildDirector


class RunModel(BaseModel):
    """pydantic model for running specifics"""
    engine: Optional[str] = 'elastic'


class ElasticModel(BaseModel):
    """elasticsearch base model for storing settings"""
    elastic_index_poi: Optional[str] = 'proplocations'
    elastic_index_users: Optional[str] = 'apiusers'
    uri: Optional[str] = 'PlacesDB:ZWFzdHVzMi5henVyZS5lbGFzdGljLWNsb3VkLmNvbTo5MjQzJGU3NDhmMjYwOTUzNTQ1NGI5ZjkxYTBjM2ViZjg0NGUyJGMyZjRiYTg1NjdkMjQ5NjE4NGJjODg4NjJlYmM5NjY5'


# ---------------------Functional Classes Here --------------------------


class CommonUtilities:
    """generic utilities used throughout entire backend script placed here"""

    # generic conversion functions including distance to radians conversion
    utility_model = common_models.CommonUtilityModel()
    feet_meter = lambda self, x: x * 0.3048
    feet_mi = lambda self, x: x / 5280
    feet_yd = lambda self, x: x / 3
    feet_km = lambda self, x: x * 0.0003048
    meter_feet = lambda self, x: x * 3.28084
    meter_km = lambda self, x: x / 1000
    km_meter = lambda self, x: x * 1000
    distance_radians = lambda self, x: x / self.utility_model.earth_radius
    mi_feet = lambda self, x: x * 5280
    proportion_prop_measurement = lambda self, x: (500 * self.radius) // x

    # generator tools/hash mapping
    generator_mapping = {'pymongo': common_models.parse_coordinates_pymongo,
                         'elastic': common_models.parse_coordinates_elastic}

    @staticmethod
    def determine_batch_size(arr: List[Any]) -> int:
        """determine batch size of concurrency operation"""

        if len(arr) == 0:
            return 0
        elif len(arr) == 1:
            return 1
        else:
            if 1 < len(arr) <= 5:
                return 3
            elif 5 < len(arr) < 10:
                return 5
            elif 10 <= len(arr) < 20:
                return 5
            else:
                return 10

    @staticmethod
    def divide_arr(arr: List[Any], n: int) -> Any:
        """divide an arr into n different subarrays"""

        k, m = divmod(len(arr), n)

        # division of array is done at this stage, multiple nested array, dim(arr)=2
        return (arr[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(n))

    @staticmethod
    def format_date(date):
        """format string dates"""

        # parse the date string here
        date_info = date.split(' ')
        final_date = date_info[0]
        time_info = date.split(' ')
        final_time = time_info[1]

        # return the final_date, final_time pair in one list obj
        return [final_date, final_time]

    @staticmethod
    def _format_address(unit, street, number, state, postcode, city, single):
        """format the address properly according to US Census regulations"""

        # if there is not unit
        if single:
            # return the singular home response here
            address = f"{number} {street}, {city}, {state}, {postcode}"
        else:
            # if there is a unit, structure differently
            address = f"{number} {street}, {city}, {state}, {postcode}, {unit}"
        return address

    @staticmethod
    def convex_hull(points):
        """calculate the convex hull of a set of points"""

        # type checker for points to avoid runtime errors
        convex_hull_model = common_models.PointsModel(coordinates=points)

        # find the hull and calculate the area
        hull = ConvexHull(convex_hull_model.coordinates)
        return hull.area

    def process_iterable(self, results: Any, first_strategy: Callable, second_strategy: Callable) -> List:
        """process generator output based on the threshold specified in the Base pydantic data model"""
        start = time.perf_counter()

        # get the length of the results
        try:
            # pymongo version
            result_len = len(list(results.clone()))
        except AttributeError:
            # elastic version
            result_len = len(results)

        if result_len > common_models.CommonUtilityModel().generator_threshold:
            # generator
            return handler_func(results, first_strategy(), self.generator_mapping[RunModel().engine], option=0)
        else:
            # iterator
            return handler_func(results, second_strategy(), self.generator_mapping[RunModel().engine], option=1)

    @staticmethod
    def clean_properties_collection(properties: dict) -> dict:
        """clean out unnecessary parts of the address properties index"""

        # remove unnecessary keys
        try:
            for key in ['hash', 'Realtor_ID', 'id', 'district']:
                del properties[key]

            # preserve the order for later iteration
            return OrderedDict(properties)
        except KeyError:
            for key in ['id', 'district']:
                del properties[key]
            return OrderedDict(properties)

    # noinspection PyTypeChecker
    def calc_distance(self, document, end: list, setting: str):
        """calculate distance between two coordinates"""

        try:
            start = document['location']['coordinates']
        except TypeError:
            start = self.generator_mapping[setting](document)

        lat1, lon1 = start
        lat2, lon2 = end

        lat1, lon1 = radians(lat1), radians(lon1)
        lat2, lon2 = radians(lat2), radians(lon2)

        dlon = lon2 - lon1
        dlat = lat2 - lat1

        a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
        c = 2 * atan2(sqrt(a), sqrt(1 - a))

        return self.utility_model.earth_radius * c


# noinspection PyCompatibility
class ManageDatabaseGeneral(BuildObjs):
    """general database management functionality, especially for code clean up"""

    def __init__(self, database: str, collections: Tuple) -> None:
        super().__init__()
        self.database_objs = DirectorGeneralObjs(database, collections, 100)

    def add_search_index(self):
        """index creation function, add all other indexes here"""
        self.database_objs[common_models.ObjModel.current_geographic_collection].create_index(
            [('location.coordinates.0', 1)])
        self.database_objs[common_models.ObjModel.current_geographic_collection].create_index(
            [('location.coordinates.1', 1)])

        # print the existing indexes
        self._manage_indexes()

    # manage the indexes of the main properties dataset
    def _manage_indexes(self):
        # get the indexes on the database
        indexes = self.database_objs[common_models.ObjModel.current_geographic_collection].list_indexes()
        # print them here for debugging purposes
        for index in list(indexes):
            print(index)


# noinspection PyCompatibility,PyUnresolvedReferences
class DensityRadius(BuildObjs, CommonUtilities):
    """density backend functionality and database querying algorithms"""

    def __init__(self, database, collection):
        super().__init__()

        # create global runner model
        self.runner_model = RunModel()
        self.global_settings = LocalDataClass()

        # call the builder/director function for SetUpObjs, make the geoCoords collection available, max_size=100
        self.database_objs = self.global_settings.builder_utility(database, collection, 100)

        # initialize basic schema for density computations
        self.density_schema = density_models.DensitySchema()

        # unit conversion algorithms collection here
        self.unit_algo_index_input: dict = dict(zip(['ft', 'mi', 'm', 'yd', 'km'],
                                                    [None, self.mi_feet, self.meter_feet, None,
                                                     *[self.km_meter, self.meter_feet]]))
        self.unit_algo_index_output: dict = dict(zip(['ft', 'mi', 'm', 'yd', 'km'],
                                                     [None, self.feet_mi, self.feet_meter, self.feet_yd, self.feet_km]))

    @staticmethod
    def compute_area(radius):
        """compute area of a given area (simple formula)"""
        area = pi * (radius ** 2)
        return area

    def _iterative_calc(self, nearest: List, radius):
        """iterative density calculator"""
        density_schema = density_models.DensitySchema(radius=radius, region_area=self.compute_area(radius))
        for _ in nearest:
            density_schema.iterative_counter += 1 / density_schema.region_area

        return density_schema.iterative_counter

    def _convert_units(self, unit_input: str, radius, **input_type):
        """Convert input units to feet to standardize"""

        # create the state machine object
        conversion_state_machine = unit_conversion_model.UnitConversionsToFeet(radius)

        # unit conversion events defined here
        unit_conversion_index_input: dict = dict(
            zip(['ft', 'mi', 'm', 'yd', 'km'], [conversion_state_machine.to_ft,
                                                conversion_state_machine.to_mi,
                                                conversion_state_machine.to_m,
                                                conversion_state_machine.to_yd,
                                                conversion_state_machine.to_km]))

        unit_conversion_index_output: dict = dict(
            zip(['ft', 'mi', 'm', 'yd', 'km'], [conversion_state_machine.to_ft_output,
                                                conversion_state_machine.to_mi_output,
                                                conversion_state_machine.to_m_output,
                                                conversion_state_machine.to_yd_output,
                                                conversion_state_machine.to_km_output]))

        # compute conversion of unit types
        try:
            # decide between input/output type
            if input_type['input'] == 'input':
                return unit_conversion_model.state_machine_conversion_controller(
                    unit_conversion_index_input[unit_input],
                    conversion_state_machine,
                    self.unit_algo_index_input[unit_input]
                )
            elif input_type['input'] == 'output':
                return unit_conversion_model.state_machine_conversion_controller(
                    unit_conversion_index_output[unit_input],
                    conversion_state_machine,
                    self.unit_algo_index_output[unit_input]
                )
        except (KeyError, BufferError):
            raise BrokenPipeError

    def run_2dSphere(self, start: List[float], radius, unit='m') -> List:
        """Handle 2dSphere queries in MongoDB for GIS object counting"""
        # using 2dSphere index

        try:
            # find connection to elasticsearch index
            elastic_connection = self.database_objs['connection']

            # create query request body
            body = {
                "size": 1000,
                "query": {
                    "bool": {
                        "filter": {
                            "geo_distance": {
                                "distance": f"{radius+self.global_settings.radius_delta_null}{unit}",
                                "location": f"{start[0]}, {start[1]}"
                            }
                        }
                    }
                },
                "sort": {
                    "_geo_distance": {
                        "location": f"{start[0]}, {start[1]}",
                        "order": "asc"
                    }
                }
            }

            try:
                nearest = elastic_connection.search(index=common_models.ObjModel().current_geographic_collection,
                                                              body=body)['hits']['hits']
            except (NotFoundError, RequestError):
                raise BrokenPipeError

        except KeyError:
            nearest = self.database_objs[common_models.ObjModel().current_geographic_collection].find({"location": {
                "$geoWithin": {
                    "$centerSphere": [start, self.distance_radians(self.feet_meter(radius))]}}})

        return nearest

    async def load_radius(self, start: List[float], radius: int) -> List:
        """load radius backend functionality"""
        try:
            nearest_poi: Any = self.run_2dSphere(start, radius)
            array = self.process_iterable(nearest_poi, generator_model.RegularGenerator,
                                          generator_model.RegularIterator)
            return array
        except OperationFailure:
            raise BrokenPipeError

    def manual_identify_density(self, coordinates: List[float], radius, unit_in: str):

        """Identify density with 2dSphere technique"""

        # convert to feet first
        radius_processed: float = self._convert_units(unit_in, radius, input='input')

        # handle radius object for density gathering
        try:
            radius_task: Any = self.run_2dSphere(coordinates, radius_processed, 'ft')
        except OperationFailure:
            raise BrokenPipeError

        # get the density comp
        return self._iterative_calc(radius_task, radius)

    def _custom_density_query(self, query_value, strategy, generator_option):
        """create the custom density query"""
        try:
            # elastic version
            elastic_connection = self.database_objs['connection']
            points, point_count = density_models.strategies_main(common_models.ObjModel().current_geographic_collection,
                                                                 query_value, strategy,
                                                                 generator_option,
                                                                 self.generator_mapping[self.runner_model.engine],
                                                                 elastic_connection)

        except (IndexError, KeyError):
            # pymongo version
            points, point_count = density_models.strategies_main(
                self.database_objs[common_models.ObjModel().current_geographic_collection], query_value, strategy,
                generator_option, self.generator_mapping[self.runner_model.engine]
            )

        return points, point_count

    def custom_density(self, strategy, query_value):
        """get density based on custom metrics such as postcode, city, and region/state"""

        # get points and point counts by inputting correct density strategy (state, postcode, city)
        points, point_count = self._custom_density_query(query_value, strategy,
                                                         generator_model.SpecializedGeneratorDN32)

        if point_count > 0:
            # convex hull results
            area = self.convex_hull(points)

            # calculate density
            density = point_count / area
            return density
        else:
            return 0

    async def density_factory(self, decision_parameter, **kwargs):
        """factory method for separating objects for custom density and generic density computations"""
        if decision_parameter == 'custom':
            # get relevant data for custom method
            strategy = kwargs['strategy']
            query_value: Any = kwargs['query_value']

            # create the coroutine object
            custom_density_result = self.custom_density(strategy, query_value)
            return custom_density_result
        else:
            # get relevant data for manual/generic density search
            coordinates: List[float] = kwargs['coordinates']
            radius: Any = kwargs['radius']
            unit_in: str = kwargs['unit_in']

            # create and handle coroutines for generic method
            generic_density_result = self.manual_identify_density(coordinates, radius, unit_in)

            return generic_density_result


# noinspection PyCompatibility,PyUnresolvedReferences
class Convex(BuildObjs, CommonUtilities):
    """convex searches implementation"""

    def __init__(self, database, collection):
        super().__init__()

        # create global settings models
        self.runner_model = RunModel()
        self.global_settings = LocalDataClass()

        # builder/director function for general objects
        self.database_objs: dict = self.global_settings.builder_utility(database, collection, 100)

    @staticmethod
    def get_polygon_area(lats, longs):
        """find the area of the polygon involved"""
        return (0.5 * np.abs(np.dot(lats, np.roll(longs, 1)) - np.dot(longs, np.roll(lats, 1)))) * 6371

    @staticmethod
    def _fix_elastic_edges(edges):
        """correct edges for elasticsearch geo-shape query purposes"""

        # create edges container
        rev_edges: list = []

        for edge in edges[0]:
            rev_edges.append(f"{edge[0]}, {edge[1]}")

        return rev_edges

    async def _polygon_search(self, edges: List) -> Any:
        """polygon/convex search"""
        try:
            try:
                # pymongo version
                query = self.database_objs[common_models.ObjModel().current_geographic_collection].find(
                    {'location': {'$geoWithin': {'$geometry': {'location.type': 'Point', "coordinates": edges}}}})
            except (KeyError, AttributeError):
                # elastic search version

                body = {
                    "size": 5000,
                    "query": {
                        "bool": {
                            "must": {
                                "match_all": {}
                            },
                            "filter": {
                                "geo_polygon": {
                                    "location": {
                                        "points": self._fix_elastic_edges(edges)
                                    }
                                }
                            }
                        }
                    }
                }

                query = self.database_objs['connection'].search(index=common_models.ObjModel().current_geographic_collection,
                                                                body=body)['hits']['hits']

            return query
        except OperationFailure:
            raise BrokenPipeError

    async def convex_director(self, edges: List) -> List:
        """handle polygon/convex search strategy"""

        # create the connection branch
        convex_task = asyncio.create_task(self._polygon_search(edges))

        try:
            results = await convex_task

            # process the results
            processed_results: list = self.process_iterable(results, generator_model.SpecializedGeneratorPN32,
                                                            generator_model.IteratorStrategyPR32)

            return processed_results

        except OperationFailure:
            raise BrokenPipeError('Edges Crossed')


"""---------------------------------Custom Exceptions----------------------------------------------------------------"""


# noinspection PyCompatibility
class CallLimitExceeded(Exception):
    """call limit exception custom class"""

    def __init__(self, batch_size: int, message: str, status_code: int) -> None:
        self.batch_size = batch_size
        self.message = message
        self.status_code = status_code
        super().__init__(message)


# noinspection PyCompatibility
class SpreadSheetLimit(Exception):
    """spreadsheet call limit class"""

    def __init__(self, batch_size: int, message: str, status_code: int) -> None:
        self.batch_size = batch_size
        self.message = message
        self.status_code = status_code
        super().__init__(message)


"""---------------------------------Forward Geocoding Functionality--------------------------------------------------"""


# noinspection PyCompatibility,PyUnresolvedReferences
class ForwardGeocoding(BuildObjs, CommonUtilities):
    """forward geocoding backend functionality"""

    def __init__(self, database, collections):
        super().__init__()

        # create background data models for forward geocoding and global settings
        self.threshold_model = LocalDataClass()
        self.runner_model = RunModel()
        self.forward_model = forward_models.ForwardGeocodingSchema(radius=self.threshold_model.optimal_radius)
        self.batch_handler = forward_models.BatchPackets()

        # create database objects with encapsulated builder pattern
        self.database_objs = self.threshold_model.builder_utility(database, collections, 100)

        # response and supply packet init here
        self.response_packet = []
        self.supplied_packet = []

        # geocoding ML model mapping
        self.forward_ml_model_mapping = {'pymongo': TFIDF_max, 'elastic': TFIDF_max_elastic}
        self.forward_coordinate_parse_mapping = {'pymongo': common_models.parse_coordinates_pymongo,
                                                 'elastic': common_models.parse_coordinates_elastic}

    @staticmethod
    def unit_check(address, address_type) -> bool:
        """check if it is a singular/multiple unit property"""

        if address_type == 1:
            unit_arr = address.split(',')
            return not not unit_arr == 5
        else:
            return not not len(address.keys()) == 5 or not not len(address.keys()) == 6

    @staticmethod
    def threading_utility(arguments, *functions) -> List[Any]:
        """provide threading functionality for arbitrary number of functions"""

        # set up max workers and submission storage array
        max_workers: int = len(functions)
        submission_arr: List[Any] = []
        output_concurrent: List[Any] = []

        # run concurrency algorithm
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            for function, arg in zip(functions, arguments):
                submission = executor.submit(function, *arg)
                submission_arr.append(submission)

            # unpack the future objects, and map to names input, .result()
            for future_obj in as_completed(submission_arr):
                output_concurrent.append(future_obj.result())

        return output_concurrent

    # noinspection PyArgumentEqualDefault
    def _similarity_TFIDF(self, street_query: str, result: BSON):
        """TFIDF similarity algorithm for geocode search"""

        return self.forward_ml_model_mapping[self.runner_model.engine].find_max_similarity(street_query, result)

    @staticmethod
    def _backsubstitute_null_parameters(original: Any, new_comparison: Any) -> Any:
        """substitute back parameters if there are nulls, with numpy arrays as inputs"""
        try:
            return null_parameters.minimize_nulls(original, new_comparison)
        except IndexError or ValueError:
            raise ValueError

    @staticmethod
    def create_dictionary(parsed_array, custom_street) -> Dict:
        """create address packet dictionary"""

        return {
            'postcode': parsed_array[1],
            'street': LocalDataClass().forward_geocode_street_method(custom_street),
            'number': parsed_array[2],
            'state': parsed_array[3],
            'unit': parsed_array[4],
            'city': parsed_array[-1]
        }

    @staticmethod
    def create_null_dictionary(original_parsed_array: Any):
        """create null address packet"""
        return {
            'postcode': original_parsed_array[1],
            'street': original_parsed_array[0],
            'number': original_parsed_array[2],
            'state': original_parsed_array[3],
            'unit': original_parsed_array[4],
            'city': original_parsed_array[-1]
        }

    def _layered_geocoder_concurrency(self, function: Callable, addresses) -> List[list]:
        """higher level concurrency method that geocodes the individual address batches"""

        # map the results to the executor based on batch_size expressed in address and radius arr sizes
        with ThreadPoolExecutor() as executor:
            executor.map(function, addresses)
        return self.batch_handler.global_batch_response_model.dict()

    def _validate_batch_size_api(self, batch_size: int, strategy: Callable) -> Any:
        if strategy is not forward_models.SpreadsheetStrategy() and batch_size > self.threshold_model.batch_size:
            raise CallLimitExceeded(
                message="API Limit Exceeded",
                status_code=505,
                batch_size=batch_size
            )

    def _validate_spreadsheet_size(self, batch_size: int, strategy: Callable) -> Any:
        if strategy is forward_models.SpreadsheetStrategy() and batch_size > self.threshold_model.spreadsheet_batch_size:
            raise SpreadSheetLimit(
                message="Spreadsheet API Limit Exceeded",
                status_code=500,
                batch_size=batch_size
            )

    def _create_new_string_parse(self, original: Any, BSON_obj: Dict) -> Any:
        """parse the new address for later comparison"""
        try:
            # parse the property details
            try:
                # for pymongo
                property_details: dict = self.clean_properties_collection(BSON_obj['properties'])
            except KeyError:
                # for elasticsearch
                property_details: dict = self.clean_properties_collection(BSON_obj['_source']['properties'])

            # unpack address components from OrderedDict
            unit, number, street, city, region, postcode = property_details.values()

        except TypeError:
            raise BrokenPipeError

        # if ValueError is encountered from back-substitution, then return the original
        try:
            return self._backsubstitute_null_parameters(
                original,
                [street, postcode, number, region, unit, city]
            )

        except ValueError:
            return [street, postcode, number, region, unit, city]

    def _process_street_generic(self, street: str) -> str:
        """remove generic subphrases from street name"""
        new_street_tokens = [word for word in street if word not in self.threshold_model.common_streets]
        return ''.join(new_street_tokens)

    def _address_general_region_query(self, number: str, postcode: str) -> Any:
        """query based only on required number and postcode parameters"""

        try: # elasticsearch
            # find connection in database object map
            elastic_connection = self.database_objs['connection']

            # create query body
            body = {
                "query": {
                    "bool": {
                        "should": [
                            {"term": {"properties.number": number}},
                            {"term": {"properties.postcode": postcode}}
                        ]
                    }
                }
            }

            # find search with multilayered -must query
            result = elastic_connection.search(index=common_models.ObjModel().current_geographic_collection, body=body)['hits']['hits']
        except (IndexError, KeyError):
            # pymongo search (and) query
            result = self.database_objs[common_models.ObjModel().current_geographic_collection].find(
                {'$and': [{'properties.number': str(number)}, {'properties.postcode': str(postcode)}]})

        return result

    def parse_addresses(self, address, **options):
        """parse address into individual pieces based on US address convention"""

        unit_check = self.unit_check(address, options['parsing'])

        # parsed just cleans up the strings and returns them below
        if options['parsing'] == 0:
            return [address['street'], address['postcode'], address['number'], address['state'], address['unit'],
                    address['city']]

        # otherwise, a parsing is needed to separate them
        elif options['parsing'] == 1:
            try:
                return forward_models.MLStrategy.parsing_strategy(address)
            except ValueError:
                return ['', '', '', '', '', '']
        else:
            return ['', '', '', '', '', '']

    # DEPRECATED
    def _legacy_geocoding(self, address: str) -> Any:
        """legacy geocoder that works with direct Regex, currently out of production"""

        parser = self.parse_addresses(address, parsing='n')
        address_parts = {'postcode': parser[3], 'number': parser[4], 'region': parser[1], 'street': parser[0],
                         'unit': parser[-1], 'city': parser[2]}

        if address_parts['unit'] == 'Singular Home':
            address_parts = {'postcode': int(parser[3]), 'number': int(parser[4]), 'region': parser[1],
                             'street': parser[0], 'unit': '', 'city': parser[2]}

        postcode = address_parts['postcode']
        number = address_parts['number']
        region = address_parts['region']
        street = address_parts['street']
        unit = address_parts['unit']
        city = address_parts['city']

        def update_street(street, index):

            components = street.split(' ')

            components.pop()

            return ' '.join(components)

        def manual_debug_queries():
            filtered_collection = self.geoCoords.aggregate([{'$match': {
                '$and': [
                    {'postcode': postcode},
                    {'number': number},
                    {'region': region},
                    {'street': street},
                    {'unit': unit}
                ]}}])

            filtered_collection = self.geoCoords.aggregate([{'$match': {
                '$and': [
                    {'postcode': postcode},
                    {'number': number},
                    {'region': region},
                    {'street': street}]
            }}])

        # ------------------------------------------------------------------------------------------

        packed_query = self.geoCoords.aggregate([{
            '$facet': {
                'Query(1)': [{'$match': {'$and': [
                    {'postcode': postcode},
                    {'number': number},
                    {'region': region},
                    {'street': street},
                    {'unit': unit}
                ]}}],
                'Query(2)': [{'$match': {'$and': [
                    {'postcode': postcode},
                    {'number': number},
                    {'region': region},
                    {'street': street}]
                }}]
            }
        }])

        packed_query = [i for i in packed_query]

        if len(packed_query[0]['Query(1)']) == 0:
            if len(packed_query[0]['Query(2)']) == 0:
                packed_collection_check = 0
                filtered_collection = 0

            elif len(packed_query[0]['Query(2)']) != 0:
                packed_collection_check = 2
                filtered_collection = 1
            else:
                packed_collection_check = 0
                filtered_collection = 0

        else:
            packed_collection_check = 1
            filtered_collection = 1

        if packed_collection_check == 0:

            index = 1
            filtered_query = []

            while filtered_collection == 0 and street != '':
                street = update_street(street, index)

                filtered_query = self.geoCoords.aggregate([{'$match': {
                    '$and': [{'postcode': postcode}, {'number': number}, {'region': region}, {'street': street}]}}])

                filtered_collection = len([i for i in filtered_query])
                index += 1

            return [i for i in filtered_query]

        else:
            return [i for i in packed_query[0]['Query({})'.format(packed_collection_check)]]

        # faceted_array.append([{'$match': {'$and': [{'postcode': postcode}, {'number': number}, {'region': region}]}}])

    def updated_geocoding(self, address: str, address_type: int) -> dict:
        """FULL: 1, PARSED: 0 ---- address type decision index"""

        # list[street, postcode, number, region, unit, city]
        original_array = self.parse_addresses(address, parsing=address_type)

        try:
            # preliminary query results here
            street, result = self.threading_utility([[original_array[0]], [original_array[2], original_array[1]]],
                                                    self._process_street_generic,
                                                    self._address_general_region_query)

            # remove null output ex. street = '': to save runtime
            if street == '':
                raise ValueError

            # similarity from TFIDF-ML algorithm
            similarity, results_bson, new_street = self._similarity_TFIDF(street, result)

            # get the new address parsed array based on empty facets
            new_array = self._create_new_string_parse(original_array, results_bson)

            # create address data packet based on similarity
            if similarity > self.threshold_model.similarity_threshold:
                street_param: str = new_street
            else:
                street_param: str = street

            # create the new address parts dict() model
            address_parts_new = self.create_dictionary(new_array, street_param)

            # get coordinates
            coordinates = self.forward_coordinate_parse_mapping[self.runner_model.engine](results_bson)

            # [coordinates, address_parts, similarity measurement, strategy]
            return forward_models.DataPacketHandler(forward_models.SuccessPacket(), address_parts_new, address,
                coordinates, similarity)

        # address formatting is done incorrectly
        except ValueError:
            address_parts_new = self.create_null_dictionary(original_array)
            return forward_models.DataPacketHandler(forward_models.EmptyGeocodingCase(), address_parts_new, address, [],
                                                    0)

    def _batch_geocoding_inner(self, addresses: List[str]) -> None:

        """secondary layer to geocoding (used for concurrency), iterates over addresses"""

        # iterate through addresses
        model = None
        for address in addresses:
            # get data from updated_geocoding
            data: Any = self.updated_geocoding(address, 1)

            # find response and supplied packets
            response, supplied = data['response_packet'], data['supplied_packet']

            # extend the forward model response/supplied packet data model
            model = self.batch_handler.BatchPacketHandler(response, supplied)

        return model

    def batch_geocode(self, geo_arr: List[Any], batch_strategy: Callable) -> Any:

        """central method to batch geocoding"""

        # get batch size
        batch_size = len(geo_arr)

        # validate the batch size and return custom HTTP exception if necessary
        try:
            self._validate_batch_size_api(batch_size, strategy=batch_strategy)
        except CallLimitExceeded:
            raise BufferError

        # condition where batch size is small to just run the iterator
        if batch_size <= self.threshold_model.small_batch_maximum:
            return self._batch_geocoding_inner(geo_arr)

        # split the addresses into appropriate packets for concurrency operation
        else:
            addresses: List[Any] = list(self.divide_arr(geo_arr, self.determine_batch_size(geo_arr)))

            # extract geocoded result from layered geocoder with cython backend
            return self._layered_geocoder_concurrency(self._batch_geocoding_inner, addresses)

    def spreadsheet_geocoding(self, geo_arr: List[Any], radius, batch_strategy: Callable) -> None:
        """spreadsheet geocoding backend feature, handles concurrency operations and output validation"""
        batch_size: int = len(geo_arr)

        try:
            self._validate_spreadsheet_size(batch_size, strategy=batch_strategy)
        except SpreadSheetLimit:
            raise BufferError

        addresses: List[Any] = list(self.divide_arr(geo_arr, self.determine_batch_size(geo_arr)))

        results: list = self._layered_geocoder_concurrency(self._batch_geocoding_inner, addresses)

        spreadsheet_c.write_csv("app/export.csv", 'w+', results)


def factory_geocode_backend_class(database: str, collection: str) -> Any:
    """factory method for creating geocoding facade"""

    obj = ForwardGeocoding(database, collection)
    # State = Enum("State", "Singular Batch Spreadsheet")
    return obj


# noinspection PyCompatibility
class ForwardGeocoderManager(metaclass=ABCMeta):
    """implement facade pattern for the geocoding options, obscure database and ML model settings from client code"""

    @abstractmethod
    def __init__(self):
        pass

    def __str__(self):
        return self.name

    @abstractmethod
    def activation_confirmation(self, request: str) -> str:
        """confirm endpoint activation function fired"""
        pass

    @abstractmethod
    def kill(self):
        """kill the endpoint activation after running cycle"""
        pass


class SingularGeocode(ForwardGeocoderManager):
    """singular endpoint response facade class"""

    def __init__(self, class_obj):
        self.geocoding_obj = class_obj
        self.name = 'Endpoint Geocoder (Regular)'

    def activation_confirmation(self, requests: str) -> str:
        """activate geocoding subprocesses in request packet"""
        return requests

    def kill(self) -> None:
        """kill processes (currently empty"""
        pass

    def process_geocoder(self, address: str, address_type: int) -> Tuple[Any, Any, Any, Any]:
        """process the singular geocoding request from client code"""
        response_data = self.geocoding_obj.updated_geocoding(address, address_type)
        return response_data


class BatchGeocode(ForwardGeocoderManager):
    """batch geocoding facade class"""

    def __init__(self, class_obj: Any):
        self.geocoding_obj = class_obj
        self.name = "Batch Geocoder (Regular)"

    def activation_confirmation(self, requests: str) -> str:
        """confirm process act."""
        return requests

    def kill(self) -> None:
        """kill processes (passed)"""
        pass

    def process_batch(self, addresses: List[str]) -> Any:
        """process the batch (common method)"""
        response_data = self.geocoding_obj.batch_geocode(addresses, forward_models.BatchEndpointStrategy())
        return response_data


class SpreadsheetGeocode(ForwardGeocoderManager):
    """spreadsheet facade class"""

    def __init__(self, class_obj):
        self.name = "Spreadsheet Geocoder (Regular)"
        self.geocoding_obj = class_obj

    def activation_confirmation(self, requests: str) -> str:
        """confirm the activation function was called"""
        return requests

    def kill(self) -> None:
        """kill the script"""
        pass

    def process_batch(self, addresses: List[str], radius) -> None:
        """process the batch for geocoding"""
        self.geocoding_obj.spreadsheet_geocoding(addresses, radius,
                                                 forward_models.SpreadsheetStrategy())


class GeneralGeocodingManager:
    """the facade class for the geocoding options"""

    def __init__(self, database: str, collection: str) -> None:
        # create the factory class object method
        self.class_obj = factory_geocode_backend_class(database, collection)

        # create the visible methods for client code
        self.single, self.batch, self.spreadsheet = SingularGeocode(self.class_obj), \
                                                    BatchGeocode(self.class_obj), \
                                                    SpreadsheetGeocode(self.class_obj)

        # get database client for frontend needs
        self.database_objs = self.class_obj.database_objs

    def start_search_process(self):
        """start the searching processes"""
        methods = [self.single.activation_confirmation("Backend Geocoder"),
                   self.batch.activation_confirmation("Backend Geocoder"),
                   self.spreadsheet.activation_confirmation("Backend Geocoder")]
        [print(i) for i in methods]

    def singular_geocode(self, address: str, address_type: int) -> Any:
        """singular geocoding functionality"""
        return self.single.process_geocoder(address, address_type)

    def batch_geocode(self, addresses: List[str]) -> Any:
        """batch geocoding functionality"""
        try:
            return self.batch.process_batch(addresses)
        except BufferError:
            raise BrokenPipeError

    def spreadsheet_geocode(self, addresses: List[str]) -> Any:
        """spreadsheet geocoding functionality"""
        self.spreadsheet.process_batch(addresses, self.class_obj.threshold_model.optimal_radius)


"""---------------------------------------Reverse Geocoding Functionality--------------------------------------------"""


# noinspection PyCompatibility
class AddressComponents(BaseModel):
    """address component model"""

    unit: Optional[str] = None
    street: Optional[str] = None
    number: Optional[Any] = None
    region: Optional[str] = None
    postcode: Optional[Any] = None
    city: Optional[str] = None


# noinspection PyCompatibility
class AddressPacket(BaseModel):
    """specific address packet to prevent clashes in frontend api"""

    address_parts: AddressComponents
    certainty: str
    format_address: str
    coordinates: List[Any]
    accuracy: float
    address_validation: bool


# noinspection PyCompatibility,PyUnresolvedReferences
class ReverseGeocoding(BuildObjs, CommonUtilities):
    """class handling reverse geocoding methods and techniques for frontend API"""

    def __init__(self, database, collection):
        super().__init__()

        # run director builder method
        self.threshold_model = LocalDataClass()
        self.runner_model = RunModel()
        self.database_objs = self.threshold_model.builder_utility(database, collection, 100)

        # create the geocoding model scheme, with radius of search 10
        self.forward_model = forward_models.ForwardGeocodingSchema(radius=self.threshold_model.optimal_radius)
        self.reverse_geocoding_array = []

    @staticmethod
    def _split_coords(coords, *args):
        if len(args) == 0:
            return coords
        else:
            return coords[::-1]

    def _create_address_packet(self, JSON: Dict) -> Tuple[dict, str]:
        property_details: dict = self.clean_properties_collection(JSON['properties'])
        model: AddressComponents = AddressComponents(**property_details)

        return model.dict(), model.unit

    @staticmethod
    def _create_threaded_radius(radius, iterations: int) -> list:
        """create radius thread"""
        return radius_arr.construct_radius_arr(radius, iterations)

    @staticmethod
    def _create_batch_inner_packet(address_components: dict, formatted_address: str,
                                   coordinate, confidence: float):

        if common_models.poi_not_found_check(formatted_address):
            accuracy: float = 0.0
            address_val: bool = False
        else:
            accuracy: float = 1.0
            address_val: bool = True

        return {'address_parts': address_components,
                'confidence': confidence,
                'coordinates': coordinate,
                'format_address': formatted_address,
                'accuracy': accuracy,
                'address_validation': address_val
                }

    # DEPRECATED
    def _address_search(self, coords: List[list]) -> Any:

        """obtain the document using reverse mapping technique, currently unused"""

        # basic query, direct coordinate
        initial_doc = self.database_objs[common_models.ObjModel().current_geographic_collection].find(
            {'location.coordinates': coords})

        geocoded_point = [i for i in initial_doc][0]
        result_length = len(geocoded_point)

        # get the unit field here
        unit_field = geocoded_point['properties']['unit']
        street = geocoded_point['properties']['street']

        try:
            number = int(geocoded_point['properties']['number'])
        except ValueError:
            number = geocoded_point['properties']['number']

        city = geocoded_point['properties']['city']
        postcode = int(geocoded_point['properties']['postcode'])
        state = geocoded_point['properties']['region']

        address_collection = {'unit': unit_field,
                              'street': street,
                              'number': number,
                              'state': state,
                              'postcode': postcode,
                              'city': city}

        # fix address based on unit existence condition
        if unit_field == '' or unit_field is None:
            return [self._format_address(geocoded_point, True), None, False, address_collection, result_length]
        else:
            return [self._format_address(geocoded_point, True), None, False, address_collection,
                    result_length]

    def _concurrent_reverse(self, function: Callable, coordinates,
                            radii) -> Iterable:

        """concurrent reverse geocoding query mapping"""

        with ThreadPoolExecutor() as executor:
            executor.map(function, coordinates, radii)

        return self.reverse_geocoding_array

    def _validate_batch_size_api(self, batch_size: int, strategy: Callable):

        """validate that the batch size was not exceeded"""

        # check batch size and existence of correct strategy, raising custom exception base class if condition met
        if strategy is reverse_models.BatchStrategy and batch_size > self.threshold_model.batch_size:
            raise CallLimitExceeded(
                message="Batch Size Limit exceeded",
                status_code=501,
                batch_size=batch_size
            )

    def _validate_spreadsheet_size(self, batch_size: int, strategy: Callable):

        """validate spreadsheet size"""

        # same as batch size validation but for spreadsheet endpoints
        if strategy() is forward_models.SpreadsheetStrategy() and batch_size > self.threshold_model.spreadsheet_batch_size:
            raise SpreadSheetLimit(
                message="Spreadsheet Size Limit Exceeded",
                status_code=501,
                batch_size=batch_size
            )

    def _coordinate_query(self, lat: float, long: float, radius=100):
        """compute coordinate query"""
        try: # elasticsearch version
            # query database dict() for elastic connection if exists
            elastic_connection = self.database_objs['connection']

            # create query body
            body = {
                "size": 10,
                "query": {
                    "bool": {
                        "filter": {
                            "geo_distance": {
                                "distance": f"{radius+self.threshold_model.radius_delta_null}m",
                                "location": f"{long}, {lat}"
                            }
                        }
                    }
                },
                "sort": {
                    "_geo_distance": {
                        "location": f"{long}, {lat}",
                        "order": "asc"
                    }
                }
            }

            # conduct elasticsearch query with desired global index setting
            try:
                nearest = elastic_connection.search(index=common_models.ObjModel().current_geographic_collection, body=body)['hits']['hits'][0]['_source']
            except (KeyError, NotFoundError, RequestError):
                raise BrokenPipeError

        except (KeyError, IndexError):
            try:
                nearest: Any = self.database_objs[common_models.ObjModel().current_geographic_collection].find(
                    {"location": {
                        "$geoWithin": {
                            "$centerSphere": [[long, lat], self.distance_radians(self.feet_meter(radius))]}}}).clone()
                try:
                    return [i for i in nearest][0]
                except IndexError:
                    # return null response without any addresses
                    return ['', AddressComponents().dict(), 0.0]

            except OperationFailure:
                raise BrokenPipeError

        return nearest

    # noinspection PyCompatibility
    def geolocation_search(self, location: List[float], radius) -> List:

        """perform reverse geocoding with 2dSphere technique"""

        # parse the coordinates here
        long, lat = self._split_coords(location)
        check_single: Callable[[Any], Any] = lambda unit_str: unit_str == '' or unit_str is None

        # get the nearest on or near that coordinate
        poi_nearest = self._coordinate_query(lat, long, radius)

        # if null
        if type(poi_nearest) is list:
            return poi_nearest

        # get confidence score
        confidence: float = certainty_models.state_controller(self.threshold_model.confidence_index,
                                                              self.calc_distance(poi_nearest,
                                                                                 location, self.runner_model.engine),
                                                              certainty_search.confidence_map)
        # get the address packet
        address_packet, unit = self._create_address_packet(poi_nearest)

        # format address in single-liner variable declaration
        # noinspection PyArgumentList
        return [self._format_address(*address_packet.values(), check_single(unit)), address_packet, confidence]

    def _reverse_inner(self, arr, geo_radius):

        """reverse inner method for computing 2dSphere operations/searches"""

        for coordinate in arr:
            formatted_address, address_components, confidence = self.geolocation_search(coordinate, geo_radius)

            # return the formatted data here with Cython/C++ support
            self.reverse_geocoding_array.append(
                self._create_batch_inner_packet(address_components, formatted_address, coordinate, confidence))

    def batch_reverse_geocoding(self, arr: list, radius: Any, strategy: Callable):

        """concurrent version of batch reverse geocoding"""

        # find the batch size (once)
        batch_size = len(arr)

        # validate batch load
        try:
            self._validate_batch_size_api(batch_size, strategy())
        except CallLimitExceeded:
            raise BufferError

        try:
            # separate plain iteration from concurrency approach
            if batch_size <= self.threshold_model.small_batch_maximum:
                self._reverse_inner(arr, self.threshold_model.optimal_radius)
                return self.reverse_geocoding_array

            else:
                # create the concurrent function inputs
                # noinspection PyCompatibility
                coordinates: List[Any] = list(self.divide_arr(arr, self.determine_batch_size(arr)))
                # noinspection PyCompatibility
                radius_arr: list = self._create_threaded_radius(radius, batch_size)

                # activate the reverse utility concurrency operator
                return self._concurrent_reverse(self._reverse_inner, coordinates, radius_arr)
        except BrokenPipeError:
            raise BrokenPipeError

    def spreadsheet_reverse(self, coordinate_arr: list, radius: Any, strategy: Callable) -> None:

        """spreadsheet reverse geocoding method"""

        # find batch load
        # noinspection PyCompatibility
        batch_size: int = len(coordinate_arr)

        # validate batch load
        try:
            self._validate_spreadsheet_size(batch_size, strategy())
        except SpreadSheetLimit:
            raise BufferError

        # extract results and return filename
        # noinspection PyCompatibility
        results: list = self.batch_reverse_geocoding(coordinate_arr, radius, strategy())
        reverse_spreadsheet.spread_write(results, 'app/export.csv', 'w')


def factory_reverse_backend(database: str, collection: str) -> Any:
    """factory implementation for reverse geocoding facade pattern"""

    back_obj = ReverseGeocoding(database, collection)
    # state = Enum("State", "Singular Batch")
    return back_obj


# noinspection PyCompatibility
class ReverseManager(metaclass=ABCMeta):
    """metaclass manager for reverse geocoding functionality"""

    @abstractmethod
    def __init__(self):
        pass

    def __str__(self):
        return self.name

    @abstractmethod
    def show_activation(self, request: str) -> None:
        """start processes"""
        pass


class SingleReverse(ReverseManager):
    """singular reverse geocoding functional exposure"""

    def __init__(self, backend_obj):
        self.reverse = backend_obj
        self.name = "Singular Reverse"

    def show_activation(self, request: str) -> None:
        """start processes here"""
        print(request)

    def single_reverse(self, location: list, radius: Any) -> Any:
        """singular geocoding implementation"""
        return self.reverse.geolocation_search(location, radius)


class BatchReverse(ReverseManager):
    """batch geocoding functional exposure"""

    def __init__(self, backend_obj):
        self.reverse = backend_obj
        self.name = "Batch Reverse"

    def show_activation(self, request: str) -> None:
        """start processes here"""
        print(request)

    def batch_reverse(self, coordinates_arr: list, radius: Any) -> Any:
        """batch revere geocoding implementation"""
        return self.reverse.batch_reverse_geocoding(coordinates_arr, radius, reverse_models.BatchStrategy)


class SpreadsheetReverse(ReverseManager):
    """spreadsheet functional exposure"""

    def __init__(self, backend_obj):
        self.reverse = backend_obj
        self.name = "Spreadsheet Reverse"

    def show_activation(self, request: str) -> None:
        """start processes"""
        print(request)

    def spread_reverse(self, coordinate_arr: list, radius: Any) -> None:
        """spreadsheet reverse geocoding implementation"""
        self.reverse.spreadsheet_reverse(coordinate_arr, radius, reverse_models.SpreadsheetStrategy)


class GeneralReverseManager:
    """general reverse geocoding coordinator class"""

    def __init__(self, database: str, collection: str) -> None:
        backend_obj = factory_reverse_backend(database, collection)
        self.single, self.batch, self.spread = SingleReverse(backend_obj), \
                                               BatchReverse(backend_obj), \
                                               SpreadsheetReverse(backend_obj)
        self.database_objs = backend_obj.database_objs

    def activate_processes(self) -> None:
        """activate the processes here using Enum from middleware layer"""
        methods = [self.single.show_activation("Activation of Singular Reverse Geocoder"),
                   self.batch.show_activation("Activation of Batch Reverse Geocoder")]
        [print(i) for i in methods]

    def single_reverse(self, location: list, radius: Any) -> Any:
        """single reverse geocoding"""
        return self.single.single_reverse(location, radius)

    def batch_reverse(self, coordinates_arr: list, radius: Any) -> Any:
        """batch reverse geocoding"""
        return self.batch.batch_reverse(coordinates_arr, radius)

    def spread_reverse(self, coordinate_arr: list, radius: Any) -> None:
        """spreadsheet reverse geocoding"""
        self.spread.spread_reverse(coordinate_arr, radius)


class Subscription:
    """contains subscription objects and functionality"""

    def __init__(self, database: str, collection: str):
        backend_obj = factory_reverse_backend(database, collection)
        self.database_objs = backend_obj.database_objs


################################ Autocompletion ########################################


class AutocompleteParse(BuildObjs, CommonUtilities):
    """autocomplete/parsing class"""

    def __init__(self, database, collection):
        super().__init__()

        # run director builder method
        self.database_objs = DirectorGeneralObjs(database, collection, 100)

        # data models
        self.settings_model = LocalDataClass()
        self.autocomplete_model = correction_parse_models.CorrectionModel()  # change this model instantiation once functionality is added
        self.unit_check = lambda unit: True if unit == '' else False

    @staticmethod
    def _run_similarity(street_query, region, city, result):
        """find most similar street"""
        return TFIDF_max.find_max_similarity(street_query, region, city, result)

    @staticmethod
    def _generator_unpacker(BSON_obj: Any) -> dict:
        for document in BSON_obj:
            yield document

    def _address_query(self, postcode: str, number: str) -> dict:
        """find general region where postcode and number filter match"""

        # process query with find() operator
        address_query = self.database_objs[common_models.ObjModel().current_geographic_collection].find(
            {'$and': [{'properties.number': number}, {'properties.postcode': postcode}]}
        )

        # unpack using generator expression
        return address_query  # self._generator_unpacker(address_query)

    def _encode_input_address(self, full_address, **encoder_option) -> Any:
        """encode the input address numerically for model integration"""
        try:
            encoder_option = encoder_option['option']
        except KeyError:
            encoder_option = self.settings_model.default_autocomplete_algorithm

    def _parse_address_network(self, encoded_full_address) -> dict:
        """parse the address into its components for parsing facade and """
        # example output: {street: ..., postcode: ..., ...}
        return {}

    def autocomplete(self, full_address: str) -> str:
        """autocompletion facade function"""

        # encode/preprocess address
        encoded_address = self._encode_input_address(full_address)

        # parse the address parts
        number, street, city, postcode, region, unit = self._parse_address_network(encoded_address)

        # regional query computing here
        regional_results = self._address_query(postcode, number)

        # get similarity results
        street_query, city_query, region_query = self._run_similarity(street, region, city, regional_results)

        # similarity from TFIDF-ML algorithm
        similarity_street, results_bson_street, new_street = street_query
        similarity_city, results_bson_city, new_city = city_query
        similarity_region, results_bson_region, new_region = region_query

        return self._format_address(unit, new_street, number, new_region, postcode, new_city, self.unit_check(unit))

    def parse_facade(self, full_address) -> dict:
        """parsing facade"""
        encoded_address = self._encode_input_address(full_address)  # select a model, else the default is used
        return self._parse_address_network(encoded_address)



"""create the radius array for batch reverse processing"""

cimport cython

ctypedef fused my_fused_type:
    cython.int
    cython.double

cpdef list construct_radius_arr(my_fused_type radius, int iterations):
    """create the radius array using a fused type"""
    return [radius for _ in range(iterations)]


import pydantic
from typing import List, Optional
from abc import ABC, abstractmethod

import sys

sys.path.insert(1, '/app/models')


# noinspection PyCompatibility
class ReverseGeocodingSchema(pydantic.BaseModel):
    """store data model for reverse geocoding functionality"""
    batch_geocode_count: Optional[int] = 0


class GeneralStrategy(ABC):
    """general strategy, regular"""
    @abstractmethod
    def iter_batch_strategy(self):
        """polymorphism using abstract base class inheritance"""
        pass


class SpreadsheetStrategy(GeneralStrategy):
    """spreadsheet strategy"""
    def iter_batch_strategy(self):
        """simple class init"""
        pass


class SingleStrategy(GeneralStrategy):
    """singular reverse, regular version"""
    def iter_batch_strategy(self):
        """regular init"""
        pass


class BatchStrategy(GeneralStrategy):
    """batch reverse strategy"""
    def iter_batch_strategy(self):
        """batch init"""
        pass



import sys
import pydantic
from typing import List, Union, Optional, Callable
from abc import ABC, abstractmethod

sys.path.insert(1, '/app')

class AutoModel(pydantic.BaseModel):
    country: Optional[str] = 'usa'

class UrlPattern(ABC):
    @abstractmethod
    def url_strategy(self, address: str, token: str, location: List[Union[float, int]], radius: Union[float, int]) -> str:
        """abstract base class for the url formation strategy pattern"""
        pass

class NullRadiusComponent(UrlPattern):
    def url_strategy(self, address: str, token: str, location: List[Union[float, int]], radius: Union[float, int]) -> str:
        url = 'https://maps.googleapis.com/maps/api/place/autocomplete/json?input={}&types=geocode&components&key={}'.format(
            address, token)
        return url

class PresentRadiusNullLocation(UrlPattern):
    def url_strategy(self, address: str, token: str, location: List[Union[float, int]], radius: Union[float, int]) -> str:
        url = 'https://maps.googleapis.com/maps/api/place/autocomplete/json?input={}&types=geocode&radius={}&components&key={}'.format(
            address, radius, token)
        return url

class OnlyRadiusPresent(UrlPattern):
    def url_strategy(self, address: str, token: str, location: List[Union[float, int]], radius: Union[float, int]) -> str:
        url = 'https://maps.googleapis.com/maps/api/place/autocomplete/json?input={}&types=geocode&location={}&components&key={}'.format(
            address, location, token)
        return url

class FillKnownsStrategy(UrlPattern):
    def url_strategy(self, address: str, token: str, location: List[Union[float, int]], radius: Union[float, int]) -> str:
        url = 'https://maps.googleapis.com/maps/api/place/autocomplete/json?input={}&types=geocode&radius={}&key={}'.format(
            address, int(radius), token)
        return url

class OrganizeUrlStrategies:
    """organize/director class for url strategy formation"""
    def __init__(self, strategy: UrlPattern):
        self.strategy = strategy

    def create_url(self, address: str, token: str, location: List[Union[float, int]], radius: Union[float, int]) -> str:
        return self.strategy.url_strategy(address=address, token=token, location=location, radius=radius)

def main_urls(address: str, token: str, location: List[Union[float, int]], radius: Union[float, int], strategy: Callable, *args) -> str:
    """return the proper url per strategy"""
    UrlObject = OrganizeUrlStrategies(strategy)
    return UrlObject.create_url(address=address, token=token, location=location, radius=radius)


from pydantic import BaseModel
from typing import Optional, Tuple
from app.models import density_models


# noinspection PyCompatibility
class GlobalDataModel(BaseModel):
    """global data class for common variables across all endpoints"""

    database: Optional[str] = 'PeekDB'
    collections: Optional[Tuple] = ("proplocations", "apiusers")
    convex_edge_limit: Optional[int] = 20
    convex_area_limit: Optional[int] = 250
    density_radius_limit: Optional[dict] = {'ft': 9843, 'm': 3000.146, 'mi': 1.864205, 'yd': 3281, 'km': 3.000146}
    density_strategy_table: Optional[dict] = {'state': density_models.StateDensityStrategy(),
                                              'city': density_models.CityDensityStrategy(),
                                              'postcode': density_models.PostcodeDensityStrategy()}
    batch_endpoint_max: Optional[int] = 125
    max_load_radius: Optional[int] = 2000



"""setup file for places geocoding package"""
from setuptools import setup, find_packages
import setuptools
from pathlib import Path
from distutils.core import setup
from distutils.extension import Extension
from Cython.Build import cythonize
import numpy as np
import numpy
import importlib.machinery
this_directory = Path(__file__).parent

VERSION = '0.4.7'
DESCRIPTION = 'Package for Places Geocoding API service'

extensions = [
    Extension(name='address_packet', sources=["address_packets.pyx"], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('certainty_search', ["certainty_searchs.pyx"], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('check_address_length', ['check_address_lengths.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('check_dups', ['check_dupss.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('check_lat_long_length', ['check_lat_long_lengths.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('coordinate_conv_package', ['coordinate_conv_packages.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('create_convex_array', ['create_convex_arrays.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('create_convex_elastic', ['create_convex_elastics.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('executor_unpack', ['executor_unpacks.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('null_parameters', ['null_parameterss.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('radius_arr', ['radius_arrs.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('reverse_spreadsheet', ['reverse_spreadsheets.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('spreadsheet_c', ['spreadsheet_cs.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('TFIDF_max', ['TFIDF_maxs.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('TFIDF_max_elastic', ['TFIDF_max_elastics.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('TFIDF_upgraded', ['TFIDF_upgradeds.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('zip_coordinate', ['zip_coordinates.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."]),
    Extension('zip_coordinates_elastic', ['zip_coordinates_elastics.pyx'], library_dirs=["."],
            include_dirs=["."], runtime_library_dirs=["."])
]

# Setting up
setup(
    name="places_geocode",
    version=VERSION,
    author="Martin Mashalov",
    author_email="hello@places.place",
    description=DESCRIPTION,
    long_description=open('/Users/martinmashalov/Documents/places_geocoding_package/README.md').read(),
    long_description_content_type="text/markdown",
    packages=['places_geocode', 'places_geocode/app', 'places_geocode/Logs', 'places_geocode/app/models', 'places_geocode/app/CythonOperations', 'places_geocode/app/dprk_copy'],
    license="MIT",
    ext_modules=cythonize(extensions, build_dir="places_geocode/app/CythonOperations"),
    #include_dirs=np.get_include(),
    url="https://medium.com/p/10cc24c34505/edit",
    install_requires=['pymongo', 'numpy', 'pandas', 'sklearn', 'dependency_injector', 'pydantic', 'fastapi', 'cython',
                      'state_machine', 'elasticsearch', 'usaddress', 'nltk'],
    keywords=['python',
              'geocoding',
              'forward geocoding',
              'reverse geocoding',
              'radius loading',
              'loading radius',
              'autocomplete',
              'places of interest',
              'POI'
              ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: Free for non-commercial use",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


from pydantic import BaseModel
from enum import Enum
from typing import Union, Optional, List
from fastapi import HTTPException, Query


# noinspection PyCompatibility
def LoadRadiusModel(class_method):
    """load properties/radius endpoint data model"""

    class LoadRadiusModelInput(str, Enum):
        parcel = "parcel"
        roof_top = "roof_top"

    class TimestampsModel(BaseModel):
        date: str = None
        official_timestamp: float = None
        time: str = None

    class InternalCoordinateArray(BaseModel):
        latitude: Optional[Union[float, int, None]] = None
        longitude: Optional[Union[float, int, None]] = None
        address: Optional[str] = None

    class GeocodedData(BaseModel):
        load_radius: Optional[List[InternalCoordinateArray]] = []
        certainty: Optional[str] = None

    class WithinMeta(BaseModel):
        build: str = None
        geocoded_data: Optional[GeocodedData] = []
        matching: int = None
        status: int = None
        timestamps: Optional[TimestampsModel] = None

    class SuppliedDataModel(BaseModel):
        latitude: float = None
        longitude: float = None
        supplied_radius: int = None

    class MetaModel(BaseModel):
        meta: Optional[WithinMeta] = None
        supplied_data: Optional[SuppliedDataModel] = []
        source: str

        class Config:
            orm_mode = True

    if class_method == 'meta':
        return MetaModel
    if class_method == 'lr':
        return LoadRadiusModelInput


# noinspection PyCompatibility
def ConvexSearchModel(class_method):
    """convex search endpoint data model"""

    class SearchModel(str, Enum):
        parcel = "parcel"
        rooftop = "roof_top"

    class TimestampsModel(BaseModel):
        date: str = None
        official_timestamp: float = None
        time: str = None

    class GeocodedDataModel(BaseModel):
        certainty: str = 'Computational Search'
        convex_array: List[dict] = []
        repeated_edges: Optional[List[list]] = []

    class WithinMeta(BaseModel):
        build: str = None
        geocoded_data: Optional[GeocodedDataModel] = {}
        matching: int = None
        status: int = None
        timestamps: Optional[TimestampsModel] = None

    class SuppliedDataModel(BaseModel):
        search_edges: Optional[List[List[float]]] = []

    class MetaModel(BaseModel):
        meta: Optional[WithinMeta] = None
        supplied_data: Optional[SuppliedDataModel] = {}
        source: str

        class Config:
            orm_mode = True

    class DataModeIn(BaseModel):
        coordinates: List[List[float]] = Query(None)

    if class_method.lower() == 'mt':
        return MetaModel
    elif class_method.lower() == 'sm':
        return SearchModel
    elif class_method.lower() == 'dm':
        return DataModeIn
    else:
        raise HTTPException(
            status_code=505,
            detail='Invalid Data Model suggested',
            headers={'Error Description': "Check the details of specified error"}
        )


# noinspection PyMissingOrEmptyDocstring,PyCompatibility
def DensitySearch(method):
    """density endpoint data model"""

    class SearchModel(str, Enum):
        TwoDSphere = "2dSphere"
        HullCentroid = "ConvexHullCentroid"

    class DensityCustomOptions(str, Enum):
        state = 'state'
        postcode = 'postcode'
        city = 'city'

    class TimestampsModel(BaseModel):
        date: str = None
        official_timestamp: float = None
        time: str = None

    class GeocodedDataModel(BaseModel):
        density: Union[float, int]
        unit: str

    class WithinMeta(BaseModel):
        build: str = None
        geocoded_data: Optional[List[GeocodedDataModel]] = []
        matching: int = None
        status: int = None
        timestamps: Optional[TimestampsModel] = None

    class SuppliedDataModel(BaseModel):
        radius: Union[float, None, int]
        coordinates: Union[List[float], None, List[None]] = None

    class MetaModel(BaseModel):
        meta: Optional[WithinMeta] = None
        supplied_data: Optional[List[SuppliedDataModel]] = []
        source: str

        class Config:
            orm_mode = True

    class DataModeIn(BaseModel):
        coordinates: List[List[float]] = Query(None)

    if method.lower() == 'mt':
        return MetaModel
    elif method.lower() == 'sm':
        return SearchModel
    elif method.lower() == 'dm':
        return DataModeIn
    elif method.lower() == 'do':
        return DensityCustomOptions
    else:
        # print('in else statement lie 933')
        raise HTTPException(
            status_code=505,
            detail='Invalid Data Model suggested',
            headers={'Error_Description': "Check the details of specified error"}
        )


# noinspection PyMissingOrEmptyDocstring,PyCompatibility
def SingleReverseModel(method):
    """singular geocoding models here"""

    class ReverseModel(str, Enum):
        rooftop = "roof_top"
        parcel_search = "parcel_search"
        centroid = "centroid"

    class TimestampsModel(BaseModel):
        date: str = None
        official_timestamp: float = None
        time: str = None

    class AddressParts(BaseModel):
        unit: Optional[str] = ''
        street: Optional[str] = ''
        number: Optional[Union[int, str]] = ''
        state: Optional[str] = ""
        postcode: Optional[Union[str, int]] = ''
        city: Optional[str] = ''

    class GeocodedDataModel(BaseModel):
        certainty: Union[str, None]
        format_address: Optional[str] = ''
        address_parts: Optional[AddressParts] = {}
        address_validation: Union[None, bool]
        confidence: Union[float, None]

    class WithinMeta(BaseModel):
        build: str = None
        geocoded_data: Optional[GeocodedDataModel] = {}
        matching: int = None
        status: int = None
        timestamps: Optional[TimestampsModel] = None

    class SuppliedDataModel(BaseModel):
        coordinates: List[float] = None
        radius: Optional[int] = None

    class MetaModel(BaseModel):
        meta: Optional[WithinMeta] = None
        supplied_data: Optional[SuppliedDataModel] = {}
        source: str

        class Config:
            orm_mode = True

    class DataModeIn(BaseModel):
        coordinates: List[List[float]] = Query(None)

    if method.lower() == 'mt':
        return MetaModel
    elif method.lower() == 'sm':
        return ReverseModel
    elif method.lower() == 'dm':
        return DataModeIn
    else:
        raise HTTPException(
            status_code=505,
            detail='Invalid Data Model suggested',
            headers={'Error_Description': "Check the details of specified error"}
        )


# noinspection PyMissingOrEmptyDocstring,PyCompatibility
def BatchReverseModel(class_method):
    """batch reverse geocoding models"""

    class ReverseMapModel(str, Enum):
        rooftop = "roof_top"
        parcel_search = "parcel_search"
        centroid = "centroid"

    class TimestampsModel(BaseModel):
        date: str = None
        official_timestamp: float = None
        time: str = None

    class AddressParts(BaseModel):
        unit: Optional[str] = None
        street: Optional[str] = None
        number: Optional[Union[int, str]] = None
        region: Optional[str] = None
        postcode: Optional[Union[int, str]] = None
        city: Optional[str] = None

    class GeocodedDataModel(BaseModel):
        format_address: Union[str, None]
        coordinates: List[float]
        address_parts: Optional[AddressParts] = []
        address_validation: Union[None, bool]
        accuracy: Union[int, float, None]

    class DuplicateModel(BaseModel):
        repeated_edges: Optional[list] = []

    class WithinMeta(BaseModel):
        build: str = None
        geocoded_data: Optional[Union[List[GeocodedDataModel], List[DuplicateModel]]] = []
        matching: int = None
        status: int = None
        timestamps: Optional[TimestampsModel] = None

    class SuppliedDataModel(BaseModel):
        coordinates: List[list] = None
        radius: Optional[int] = None

    class MetaModel(BaseModel):
        meta: Optional[WithinMeta] = None
        supplied_data: Optional[SuppliedDataModel] = None
        source: str

        class Config:
            orm_mode = True

    class DataModeIn(BaseModel):
        coordinates: List[List[float]] = Query(None)

    if class_method.lower() == 'mt':
        return MetaModel
    elif class_method.lower() == 'sm':
        return ReverseMapModel
    elif class_method.lower() == 'dm':
        return DataModeIn
    else:
        # print('in else statement lie 933')
        raise HTTPException(
            status_code=505,
            detail='Invalid Data Model suggested',
            headers={'Error_Description': "Check the details of specified error"}
        )


# noinspection PyCompatibility
def SingleGeocodeModel(class_method):
    """singular geocoding response model and more"""

    class ReverseMapModel(str, Enum):
        rooftop = "roof_top"
        parcel_search = "parcel_search"
        centroid = "centroid"

    class TimestampsModel(BaseModel):
        date: str
        official_timestamp: float
        time: str

    class GeocodedDataModel(BaseModel):
        coordinates: Optional[List[Union[float, None]]] = [None, None]
        accuracy: Union[float, None, int]
        address: Optional[Union[str, None, dict]] = None
        address_parts: Optional[dict] = {}

    class WithinMeta(BaseModel):
        build: str
        geocoded_data: Optional[List[GeocodedDataModel]] = []
        matching: int
        status: int
        timestamps: Optional[TimestampsModel] = None

    class AddressPartsModel(BaseModel):
        street: Union[str, None]
        number: Union[int, str, None]
        postcode: Union[int, str, None]
        state: Union[str, None]
        city: Union[str, None]
        unit: Union[str, None]

    class SuppliedDataModel(BaseModel):
        address: Optional[Union[AddressPartsModel, str]] = None

    class MetaModel(BaseModel):
        meta: Optional[WithinMeta] = None
        supplied_data: Optional[List[SuppliedDataModel]] = None
        source: str

        class Config:
            orm_mode = True

    class DataModeIn(BaseModel):
        coordinates: List[List[float]] = Query(None)

    if class_method.lower() == 'mt':
        return MetaModel
    elif class_method.lower() == 'sm':
        return ReverseMapModel
    elif class_method.lower() == 'dm':
        return DataModeIn
    else:
        raise HTTPException(
            status_code=505,
            detail='Invalid Data Model suggested',
            headers={'Error_Description': "Check the details of specified error"}
        )


# noinspection PyCompatibility
def BatchGeocodeModel(class_method):
    """batch geocoding models"""

    class ReverseMapModel(str, Enum):
        rooftop = "roof_top"
        parcel_search = "parcel_search"
        centroid = "centroid"

    class TimestampsModel(BaseModel):
        date: str = None
        official_timestamp: float = None
        time: str = None

    class AddressPartsModel(BaseModel):
        street: Union[str, None]
        number: Union[str, int, None]
        postcode: Union[str, int, None]
        state: Union[str, None]
        city: Union[str, None]
        unit: Union[str, None]

    class GeocodedDataModel(BaseModel):
        coordinates: List[Union[float, None]] = None
        address_parts: Optional[Union[AddressPartsModel, str]] = None
        address: Optional[str] = None
        accuracy: Optional[float] = None

    class WithinMeta(BaseModel):
        build: str = None
        geocoded_data: Optional[List[GeocodedDataModel]] = None
        matching: int = None
        status: int = None
        timestamps: Optional[TimestampsModel] = None

    class SuppliedDataModel(BaseModel):
        address: Optional[AddressPartsModel] = None
        formatted_address: Union[str, None]

    class MetaModel(BaseModel):
        meta: Optional[WithinMeta] = None
        supplied_data: Optional[List[SuppliedDataModel]] = None
        source: str

        class Config:
            orm_mode = True

    class DataModeIn(BaseModel):
        coordinates: List[List[float]] = Query(None)

    if class_method.lower() == 'mt':
        return MetaModel
    elif class_method.lower() == 'sm':
        return ReverseMapModel
    elif class_method.lower() == 'dm':
        return DataModeIn
    else:
        raise HTTPException(
            status_code=505,
            detail='Invalid Data Model suggested',
            headers={'Error_Description': "Check the details of specified error"}
        )


# noinspection PyCompatibility
class Subscription(BaseModel):
    """subscription frontend model"""
    token: str


# noinspection PyCompatibility
def Autocompletion(class_method):
    """singular geocoding response model and more"""

    class TimestampsModel(BaseModel):
        """timestamps model"""
        date: str
        official_timestamp: float
        time: str

    class AutocompletedDataModel(BaseModel):
        """geocoded data model"""
        accuracy: Union[float, None, int]
        corrected_address: Optional[Union[str, None, dict]] = None

    class WithinMeta(BaseModel):
        """under meta data model"""
        build: str
        geocoded_data: Optional[AutocompletedDataModel] = {}
        matching: int
        status: int
        timestamps: Optional[TimestampsModel] = None

    class AddressPartsModel(BaseModel):
        """address component model"""
        street: Union[str, None]
        number: Union[int, None]
        postcode: Union[int, None]
        state: Union[str, None]
        city: Union[str, None]
        unit: Union[str, None]

    class SuppliedDataModel(BaseModel):
        """supplied data"""
        full_address: Optional[Union[AddressPartsModel, str]] = None

    class MetaModel(BaseModel):
        """meta model"""
        meta: Optional[WithinMeta] = None
        supplied_data: Optional[List[SuppliedDataModel]] = None
        source: str

        class Config:
            """enable orm mode"""
            orm_mode = True

    if class_method.lower() == 'mt':
        return MetaModel
    elif class_method.lower() == 'sm':
        return AutocompletedDataModel
    else:
        raise HTTPException(
            status_code=505,
            detail='Invalid Data Model suggested',
            headers={'Error_Description': "Check the details of specified error"}
        )


# noinspection PyCompatibility
def AddressParse(class_method):
    """singular geocoding response model and more"""

    class TimestampsModel(BaseModel):
        """timestamps model"""
        date: str
        official_timestamp: float
        time: str

    class ParsedDataModel(BaseModel):
        """geocoded data model"""
        accuracy: Union[float, None, int]
        address_parts: Optional[dict] = {}

    class WithinMeta(BaseModel):
        """under meta data model"""
        build: str
        geocoded_data: Optional[ParsedDataModel] = {}
        matching: int
        status: int
        timestamps: Optional[TimestampsModel] = None

    class AddressPartsModel(BaseModel):
        """address component model"""
        street: Union[str, None]
        number: Union[int, None]
        postcode: Union[int, None]
        state: Union[str, None]
        city: Union[str, None]
        unit: Union[str, None]

    class SuppliedDataModel(BaseModel):
        """supplied data"""
        full_address: Optional[Union[AddressPartsModel, str]] = None

    class MetaModel(BaseModel):
        """meta model"""
        meta: Optional[WithinMeta] = None
        supplied_data: Optional[List[SuppliedDataModel]] = None
        source: str

        class Config:
            """enable orm mode"""
            orm_mode = True

    if class_method.lower() == 'mt':
        return MetaModel
    elif class_method.lower() == 'sm':
        return ParsedDataModel
    else:
        raise HTTPException(
            status_code=505,
            detail='Invalid Data Model suggested',
            headers={'Error_Description': "Check the details of specified error"}
        )





"""address autocomplete/correction and address parsing models go here"""

from pydantic import BaseModel
from typing import List, Union, Tuple, Dict


class CorrectionModel(BaseModel):
    """dummy correction model"""
    pass


import csv

cpdef spread_write(list arr, str filename, str mode):
    cdef const char *fieldnames[3]
    fieldnames[:9] = ['latitude', 'longitude', 'address', 'number', 'street', 'city', 'postcode', 'region', 'validation']
    cdef float latitude = 0.0
    cdef float longitude = 0.0
    cdef dict packet = {}
    cdef str formatted_address = ''
    cdef bint address_val = True
    cdef dict row_keys = {}

    with open(filename, mode) as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()

        for packet in arr:
            formatted_address = packet['formatted_address']
            latitude, longitude = packet['coordinate'][0], packet['coordinate'][1]
            row_keys = {'address': formatted_address, 'latitude': latitude, 'longitude': longitude}
            row_keys.extend(packet['address_parts'])

            writer.writerow(row_keys)


cpdef list convex_array(list latitudes, list longitudes, bint reverse_param):

    """

    :param latitudes: 
    :param longitudes: 
    :param reverse_param: 
    :return: zipped coordinates
    
    """

    cdef list final_coordinates = []
    cdef list coordinates = []

    if reverse_param:
        coordinates = [[long, lat] for lat, long in zip(latitudes, longitudes)]
    else:
        coordinates = [[lat, long] for lat, long in zip(latitudes, longitudes)]

    coordinates.append(coordinates[0])
    final_coordinates.append(coordinates)

    return final_coordinates


"""handle null parameters and maximize accuracy"""
cpdef minimize_nulls(list original, list new_arr):

    """minimize the nulls present in final array with stack data structure"""

    cdef list optimized_original = [original[0]]
    cdef int i = 0

    for i in range(1, len(original)):
        if new_arr[i] == '':
            optimized_original.append(original[i])
        else:
            optimized_original.append(new_arr[i])

    return optimized_original


import logging
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Union
import pydantic


class LoggingSchema(pydantic.BaseModel):
    """dataclass for Logger Schema containing main functions"""
    level: Union[str, None]
    message: Union[str, None]
    formatter: Any
    handler: Any
    logger: Any


class LoggingData:

    """logging data main function"""

    def __init__(self, name, file, level, message):
        self.level = level
        self.message = message
        self.formatter = self._create_formatting()
        self.handler = self._create_handler(self.formatter, file)
        self.logger = logging.getLogger(name)

        # check if logger has handlers, avoids exceptions
        if not self.logger.hasHandlers():
            self.logger.addHandler(self.handler)

        # create the LoggingSchema model and convert to dictionary type
        self.model = LoggingSchema(level=self.level,
                                   message=self.message,
                                   formatter=self.formatter,
                                   handler=self.handler,
                                   logger=self.logger).dict()

    @staticmethod
    def _create_formatting():
        """create the logger formatter object"""
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
        return formatter

    @staticmethod
    def _create_handler(formatter, file):
        """create the handler object for the logger process"""
        handler = logging.FileHandler(file)
        handler.setFormatter(formatter)
        return handler


class LoggingStrategy(ABC):
    """abstract base class for the general logging strategy"""

    @abstractmethod
    def create_strategy(self, logging_data: Dict) -> None:
        """create strategy for abstract method"""
        pass


class GeneralStrategy(LoggingStrategy):
    """general strategy for adding generic logging statements"""

    def create_strategy(self, logging_data: Dict) -> None:
        """create strategy for general logging"""
        logging_data['logger'].setLevel(logging_data['level'])
        logging_data['logger'].info(logging_data['message'])


class LogError:

    """log the error encountered during API Runtime"""

    def __init__(self, logging_strategy: LoggingStrategy):
        self.logging_strategy = logging_strategy

    def do_logging(self, name, file, level, message):
        """perform logging by implementing LoggingSchema dataclass and abstract base class"""
        logging_obj = LoggingData(name, file, level, message)
        model = logging_obj.model

        # perform logging
        try:
            self.logging_strategy.create_strategy(model)
        except BaseException:
            raise Exception('Error in Logging Module Occurred')


def log_error(name, file, level, message):
    """log error client function"""
    logger = LogError(GeneralStrategy())
    logger.do_logging(name, file, level, message)

# log_error('AddUserLogger', '/Users/martinmashalov/Documents/Python/PeekGeo/Logs/LoadPropertiesLogs.txt', "DEBUG", 'test message here')



from pydantic import BaseModel, validator
from typing import List, Optional, Union, Type, Tuple, Any, Callable, Dict
from abc import ABC, abstractmethod
import sys
import nltk
#nltk.download('stopwords')
#nltk.download('wordnet')
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
import threading
# from keras.preprocessing.sequence import pad_sequences
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline
import re
import joblib
from time import perf_counter
from concurrent.futures import ThreadPoolExecutor, as_completed
from difflib import SequenceMatcher
from pickle import dump, load, HIGHEST_PROTOCOL
import json
import usaddress


# global utility
def clean_street(street):
    """clean out common subphrases that confuse similarity ML model"""
    refined_street_string = ''

    # expect parsed input of street
    for word in street:
        # check in the dataclass model for the common word match
        if word.lower() not in ForwardGeocodingSchema.common_streets:
            if street.index(word) == 0:
                refined_street_string += word
            else:
                refined_street_string += " " + word
        else:
            continue

    # give the refined street string without the common words or such
    return refined_street_string


# noinspection PyCompatibility,PyMissingOrEmptyDocstring
class AddressPartsModel(BaseModel):
    street: Union[str, None]
    unit: Optional[str] = ''
    number: Union[str, int]
    postcode: Union[str, int]
    region: Union[str, None]
    city: Union[str, None]


# noinspection PyCompatibility
class ForwardGeocodingSchema(BaseModel):
    """store data model for forward geocoding functionality"""

    earth_radius: Optional[Union[float, int]] = 6378.1 * 1000
    radius: Union[List, List[List], int, float]
    common_streets: Optional[List[str]] = ['road', 'street', 'avenue', 'drive', 'lane', 'route', 'court', 'west',
                                           'east', 'north',
                                           'south', 'place', 'state', 'boulevard', 'county', 'way', 'square']
    batch_spreadsheet_count: Optional[int] = 0
    stopwords_index: Optional[str] = 'english'


"""-------------------------------Pydantic Error Models ---------------------------------"""


# noinspection PyCompatibility
class IncorrectLengthError(Exception):
    """error class in case parsing length is too short"""

    def __init__(self, length: int, message: str, status_code: int) -> None:
        self.length = length
        self.message = message
        self.status_code = status_code
        super().__init__(message)


# noinspection PyCompatibility
class UnitCheckError(Exception):
    """custom exception class for unit checking errors"""

    def __init__(self, unit_check: bool, message: str, status_code: int) -> None:
        self.unit_check = unit_check
        self.message = message
        self.status_code = status_code
        super().__init__(message)


""""-------------------------------parsing models ----------------------------------------"""


# noinspection PyCompatibility,PyMissingOrEmptyDocstring
class AddressParseModel(BaseModel):
    street: Optional[str]
    postcode: Optional[Union[int, str]]
    number: Optional[Union[int, str]]
    region: Optional[str]
    unit: Optional[Union[None, str]]
    city: Optional[str]

    decode_vocabulary = {1: 'a', 2: 'b', 3: 'c', 4: 'd', 5: 'e', 6: 'f', 7: 'g', 8: 'h', 9: 'i', 10: 'j', 11: 'k',
                       12: 'l', 13: 'm', 14: 'n', 15: 'o',
                       16: 'p', 17: 'q', 18: 'r', 19: 's', 20: 't', 21: 'u', 22: 'v', 23: 'w', 24: 'x', 25: 'y',
                       26: 'z', 27: '1', 28: '2',
                       29: '3', 30: '4', 31: '5', 32: '6', 33: '7', 34: '8', 35: '9', 36: '0', 37: '-', 38: ' ',
                       39: '/', 40: "'"}
    vocabulary = {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5, 'f': 6, 'g': 7, 'h': 8, 'i': 9, 'j': 10, 'k': 11, 'l': 12, 'm': 13, 'n': 14, 'o': 15, 'p': 16, 'q': 17, 'r': 18, 's': 19, 't': 20, 'u': 21, 'v': 22, 'w': 23, 'x': 24, 'y': 25, 'z': 26, '1': 27, '2': 28, '3': 29, '4': 30, '5': 31, '6': 32, '7': 33, '8': 34, '9': 35, '0': 36, '-': 37, ' ': 38, '/': 39, "'": 40}


# noinspection PyCompatibility
class SimilarityAddressParser(BaseModel):
    """base model for new similarity parser"""
    state_new: str
    postcode_new: str
    number_new: str
    city_new: str
    unit_new: str


"""-------------------------------parsing strategy-----------------------------------------"""


class ParsingStrategy(ABC):
    @abstractmethod
    def parsing_strategy(self, address: str) -> str:
        """"address parsing models here"""
        pass


class StreetStrategy(ParsingStrategy):
    def parsing_strategy(self, address: str) -> str:
        street = (address.split(',')[0]).split(' ')[1:]
        return ' '.join(street)


class PostcodeStrategy(ParsingStrategy):
    def parsing_strategy(self, address: str, *args) -> str:
        parsed_address = address.split(',')[3]
        model_in = ''.join(list(parsed_address)[1:])
        return model_in


class NumberStrategy(ParsingStrategy):
    def parsing_strategy(self, address: str, *args) -> str:
        parsed_number = address.split(',')[0].split(' ')[0]
        return parsed_number


class UnitStrategy(ParsingStrategy):
    def parsing_strategy(self, address: str, *args) -> str:
        unit_check = args[0]
        try:
            if unit_check:
                parsed_address = address.split(',')[-1]
                model_in = ''.join(list(parsed_address)[1:])
                return model_in
            else:
                return ''
        except IndexError or TypeError:
            raise UnitCheckError


class RegionStrategy(ParsingStrategy):
    def parsing_strategy(self, address: str, *args) -> str:
        region = address.split(',')[2]
        model_in = ''.join(list(region[1:]))
        return model_in


class CityStrategy(ParsingStrategy):
    def parsing_strategy(self, address: str, *args) -> str:
        model_in = ''.join(list(address.split(',')[1])[1:])
        return model_in


def strategy_builder(address: str, *args) -> list:
    """director function for organizing parsing strategies"""

    # get the address length
    address_length = list(address).count(',')

    # create the pydantic local dataclass
    try:
        parse_model = AddressParseModel(address_length=address_length)

        # builder steps
        parse_model.street = StreetStrategy().parsing_strategy(address)
        parse_model.number = int(NumberStrategy().parsing_strategy(address))
        parse_model.postcode = int(PostcodeStrategy().parsing_strategy(address))
        parse_model.region = RegionStrategy().parsing_strategy(address)
        parse_model.city = CityStrategy().parsing_strategy(address)

        try:
            parse_model.unit = UnitStrategy().parsing_strategy(address, args[0])
        except UnitCheckError:
            raise ValueError

        # return model as dictionary type
        return list(parse_model.dict().values())

    except IncorrectLengthError or UnitCheckError:
        raise ValueError


"""-------------------------------------- DEEP LEARNING PARSER----------------------------------------"""


class Lemmatize(BaseEstimator, TransformerMixin):

    """lemmatize the input text"""

    def __init__(self, lemmatize=False):
        self.lemmatize = lemmatize
        self.stop_words = stopwords.words("english")
        self.wnl = WordNetLemmatizer()

    def lemmatizer_func(self, text, cores=3):
        """lemmatize text with multiprocessing pool"""
        with Pool(processes=cores) as pool:
            result = pool.map(self.wnl.lemmatize, text)
        return result

    def fit(self, address):
        return self

    def transform(self, address):
        # lemmatize the text to remove stopwords
        text = re.sub("@\S+|https?:\S+|http?:\S|[^A-Za-z0-9]+", ' ', str(address).lower()).strip()
        if self.lemmatize:
            start3 = perf_counter()
            text = ' '.join([self.wnl.lemmatize(i) for i in text.split(' ') if i not in self.stop_words])
            print(perf_counter() - start3, 'lemmatize time')

        return text


class RemoveStops(BaseEstimator, TransformerMixin):

    """remove symbols and OOVs from input vector"""

    def __init__(self, remove_spaces=False):
        self.remove_spaces = remove_spaces
        # create the vocabulary embedding index
        self.vocabulary = AddressParseModel().vocabulary
        self.vocab_set = set(self.vocabulary.keys())

    def fit(self, address):
        return self

    def transform(self, address):
        # convert address string to set
        address_set = set(address)

        # find different characters
        char_diff = address_set.difference(self.vocab_set)

        # remove those char stopages from input string
        for char in char_diff:
            address = address.replace(char, '')

        return address


class CharEmbedding(BaseEstimator, TransformerMixin):

    """embed the input address string with either one-hot encoding or custom label autoencoder"""

    def __init__(self, one_hot=False, label_encode=True):
        super().__init__()
        self.one_hot = one_hot
        self.label_encode = label_encode
        self.vocabulary = AddressParseModel().vocabulary

    def label_encoder_labels(self, word):
        vectors = []
        for w in word:
            vector = [self.vocabulary[char] for char in list(w)]
            vectors.append(vector)
        return vectors

    def fit(self, address):
        return self

    def transform(self, address):
        # perform char embedding autoencoding on the input address string
        addresses = pad_sequences(self.label_encoder_labels([address]), padding="post", maxlen=73)  # pad sequences

        # convert to list
        address_pre_arr = [list(address) for address in addresses]

        # translate to numpy array
        return np.array(address_pre_arr)


class PipelineModel(BaseModel):
    full_pipe = Pipeline([
        ("lemmatize", Lemmatize()),
        ("symb", RemoveStops()),
        ("embedchar", CharEmbedding())
    ])

    class Config:
        def __init__(self):
            pass

        arbitrary_types_allowed = True


class CustomPredictFull:

    """custom prediction method for all component models"""

    def __init__(self, address):
        self.address = address
        self.cleaned_address = None
        self.pipeline = PipelineModel().full_pipe
        self.decode_vocab = AddressParseModel().decode_vocabulary
        self.finished_predicts = {}

    def decode_sequence(self, ex_seq):
        """decode the numpy address sequences"""
        address = ''
        for i in ex_seq:
            if i != 0:
                address += self.decode_vocab[i]
        return address

    @staticmethod
    def make_prediction(address: Any):
        """make prediction on address"""
        return usaddress.parse(address)

    def predict(self):
        """make predictions for each component class in address"""
        # make prediction
        prediction = self.make_prediction(self.address)

        # temporary vars
        components = {"street": '', "postcode": '', "number": '', "region": '', "unit": '', "city": ''}

        # create address component packet
        for comp_tup in prediction:
            if comp_tup[1] == "AddressNumber":
                sub_comp = comp_tup[0].replace(",", "")
                sub_comp = sub_comp.replace(" ", "")
                components['number'] += f'{sub_comp}'
            elif "StreetName" in comp_tup[1]:
                sub_comp = comp_tup[0].replace(",", "")
                sub_comp = sub_comp.replace(" ", "")
                components['street'] += f'{sub_comp} '
            elif "PlaceName" in comp_tup[1]:
                sub_comp = comp_tup[0].replace(",", "")
                sub_comp = sub_comp.replace(" ", "")
                components['city'] += f'{sub_comp} '
            elif comp_tup[1] == 'StateName':
                sub_comp = comp_tup[0].replace(",", "")
                sub_comp = sub_comp.replace(" ", "")
                components['region'] += f'{sub_comp}'
            elif comp_tup[1] == 'AddressNumberSuffix' or comp_tup[1] == 'AddressNumberPrefix':
                sub_comp = comp_tup[0].replace(",", "")
                sub_comp = sub_comp.replace(" ", "")
                components['unit'] += f'{sub_comp}'
            elif comp_tup[1] == 'ZipCode':
                sub_comp = comp_tup[0].replace(",", "")
                sub_comp = sub_comp.replace(" ", "")
                components['postcode'] += f'{sub_comp}'

        return list(components.values())

    def correct_predictions(self):
        """correct the predictions by TFIDF text similarity"""

        self.finished_predicts = self.predict()

        for prediction in self.finished_predicts.keys():
            # get length of prediction and set up similarity index
            pred_len: int = len(self.finished_predicts[prediction].split(' '))

            similarity_scores: list = []

            # split original address into n parts where n = pred_len
            address_arr: list = self.cleaned_address.split(' ')

            # iterate through address arr
            for i in range(0, len(address_arr)-pred_len):
                # comparison strings
                comparison_comps = ' '.join(address_arr[i:i+pred_len])

                # find similarity and append tuple packet
                similarity = float(SequenceMatcher(None, comparison_comps, self.finished_predicts[prediction]).ratio())
                similarity_scores.append((similarity, comparison_comps))

            # get the most similar string
            similarity_scores.sort()
            self.finished_predicts[prediction] = similarity_scores[-1][1]

        return self.finished_predicts


class LearnParse(ABC):
    """deep learning strategy base model"""

    @abstractmethod
    def parsing_strategy(self, address: str) -> str:
        """"address parsing models here"""
        pass


class MLStrategy(LearnParse):

    @staticmethod
    def parsing_strategy(address: str) -> dict:
        """call predict method in prediction class"""
        strat_obj = CustomPredictFull(address)
        return strat_obj.predict()


"""-------------------------------------- Response Packet Strategy ------------------------------------"""


# noinspection PyCompatibility
class ResponsePacketDataClass(BaseModel):
    """sub-response storage data model"""
    response_packet: Optional[List] = []
    supplied_packet: Optional[List] = []


class ResponseFormStrategy(ABC):
    """response form strategy abstract base class"""

    @abstractmethod
    def response_packet_strategy(self, address_parts: Dict, address: Union[str, None],
                                 coordinates: List[Union[int, float]], accuracy: Union[int, float]) -> dict:
        """define response packet polymorphism method"""
        pass


class SuccessPacket(ResponseFormStrategy):
    """successful return packet"""

    def response_packet_strategy(self, address_parts: Dict, address: Union[str, None],
                                 coordinates: List[Union[int, float]], accuracy: Union[int, float]) -> dict:
        """
        :param address_parts:
        :param address:
        :param coordinates:
        :param accuracy:
        """

        local_response_model = ResponsePacketDataClass()

        response = {'address_parts': address_parts, 'address': address,
                    'coordinates': coordinates, 'accuracy': accuracy}
        supplied = {'address': address_parts, 'formatted_address': address}

        local_response_model.response_packet.append(response)
        local_response_model.supplied_packet.append(supplied)

        return local_response_model.dict()


class EmptyGeocodingCase(ResponseFormStrategy):
    """empty forward geocoding (no query output)"""

    def response_packet_strategy(self, address_parts: Dict, address: Union[str, None],
                                 coordinates: List[Union[int, float]], accuracy: Union[int, float]) -> dict:
        """create an empty geocode response for error cases"""

        local_response_model = ResponsePacketDataClass()

        response = {'address_parts': address_parts, 'address': address,
                    'coordinates': [], 'accuracy': accuracy}
        supplied = {'address': address_parts, 'formatted_address': address}

        local_response_model.response_packet.append(response)
        local_response_model.supplied_packet.append(supplied)

        return local_response_model.dict()


class ErrorCase(ResponseFormStrategy):
    """error case (query returns 505 internal error"""

    def response_packet_strategy(self, address_parts: Dict, address: Union[str, None],
                                 coordinates: List[Union[int, float]], accuracy: Union[int, float]) -> dict:
        """Internal Server Error response cases"""

        local_response_model = ResponsePacketDataClass()

        response = {'address_parts': address_parts, 'address': 'Address Formatting Error',
                    'coordinates': [], 'accuracy': accuracy}
        supplied = {'address': address_parts, 'formatted_address': address}

        local_response_model.response_packet.append(response)
        local_response_model.supplied_packet.append(supplied)

        return local_response_model.dict()


def DataPacketHandler(strategy: ResponseFormStrategy, address_parts: Dict, address: Union[str, None],
                      coordinates: List[Union[int, float]], accuracy: Union[int, float]) -> dict:
    """handle data packet creation"""
    return strategy.response_packet_strategy(address_parts, address, coordinates, accuracy)


class BatchPackets:

    """handle all the batch packaging for batch endpoints"""

    def __init__(self):
        self.global_batch_response_model = ResponsePacketDataClass()

    def BatchPacketHandler(self, response_packet: Union[List, None], supplied_packet: Union[List, None]) -> dict:
        """handle the batch packets and extend to response/supplied models"""

        self.global_batch_response_model.response_packet.extend(response_packet)
        self.global_batch_response_model.supplied_packet.extend(supplied_packet)

        return self.global_batch_response_model.dict()


"""---------------------------------------Spreadsheet Batch Strategy------------------------------------------------"""


class BatchStrategy(ABC):
    """batch strategy abstract base class"""

    @abstractmethod
    def batch_strategy(self) -> None:
        """polymorphism method init"""
        pass


class SpreadsheetStrategy(BatchStrategy):
    """spreadsheet"""

    def batch_strategy(self) -> None:
        """batch strategy func"""
        pass


class BatchEndpointStrategy(BatchStrategy):
    """batch"""

    def batch_strategy(self) -> None:
        """batch strategy func"""
        pass


class SingularEndpointStrategy(BatchStrategy):
    """regular/singular"""

    def batch_strategy(self) -> None:
        """batch strategy func"""
        pass
